COUNTRIES = (("countries",),{'type':str,'nargs':'*','help':"countries to work on",'metavar':'COUNTRY'})
OVERPASS = (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL",'metavar':'URL'})
NOGT = (("-g","--no-graph-tool"),{'action':'store_true','dest':'no_gt','default':False,'help':"do not perform graph_tool tests"})
URL = (("-d","--download-url"),{'type':str,'dest':'url','help':"define url for download directory to be URL",'metavar':'URL'})
SHOW = (("-l","--list"),{'action':'store_true','dest':'show','default':False,'help':"list available countries (default: False)"})
CONSERVEMEM = (("-c","--conserve-memory"),{'action':'store_true','dest':'conserve_mem','default':False,'help':"lower memory usage but higher runtime"})
CONFIG = [
    ("stage0",{'help':"generate map data and all graph files (VERY SLOW)",'args':[SHOW,OVERPASS,NOGT,URL,COUNTRIES]}),
    ("stage1",{'help':"download map data and generate all graph files (SLOW)",'args':[
        ("cache",{'help':"cache files from Overpass API",'args':[SHOW,OVERPASS,NOGT,URL,COUNTRIES]}),
        ("osm",{'help':"compressed XML from GeoFabrik",'args':[SHOW,CONSERVEMEM,NOGT,URL,COUNTRIES]})
    ]}),
    ("stage2",{'help':"download NX graph files and other graph files",'args':[SHOW,NOGT,URL,COUNTRIES]}),
    ("stage3",{'help':"download all graph files (RECOMMENDED)",'args':[SHOW,NOGT,URL,COUNTRIES]})
]

def fill_all(overpass_url,countries):
    from limic.fill import fill
    for country in countries:
        file_name = "cache."+country
        fill(overpass_url,file_name=file_name,area=None,around=1000,eps=0.01,safe_dist=100,penalize=20,max_workers=None)

def extract_cache_all(overpass_url,countries):
    from limic.extract import extract_cache
    for country in countries:
        file_name_in = "cache."+country
        file_name_out = "graph."+country+".nx"
        extract_cache(file_name_in,file_name_out,overpass_url,area=None,around=1000,eps=0.01,safe_dist=100,penalize=20)

def extract_osm_all(countries,conserve_mem):
    from limic.extract import extract_osm
    for country in countries:
        file_name_in = country+"-latest.osm.bz2"
        file_name_out = "graph."+country+".nx"
        extract_osm(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20,conserve_mem=conserve_mem)

def convert_nx_gt_all(countries):
    from limic.convert import convert_nx_gt
    for country in countries:
        file_name_in = "graph."+country+".nx"
        file_name_out = "graph."+country+".gt"
        convert_nx_gt(file_name_in,file_name_out)

def convert_nx_npz_all(countries):
    from limic.convert import convert_nx_npz
    for country in countries:
        file_name_in = "graph."+country+".nx"
        file_name_out = "graph."+country+".npz"
        convert_nx_npz(file_name_in,file_name_out)

def convert_gt_npz_all(countries):
    from limic.convert import convert_gt_npz
    for country in countries:
        file_name_in = "graph."+country+".gt"
        file_name_out = "graph."+country+".npz"
        convert_gt_npz(file_name_in,file_name_out)

def merge_all(countries):
    from limic.merge import merge_nx
    from limic.convert import convert_nx_npz
    file_names = list(map(lambda country:"graph."+country+".nx",countries))
    merge_nx(file_names,"merged.Europe.nx")
    convert_nx_npz("merged.Europe.nx","merged.Europe.npz")

def convert_merge_all(countries,no_gt):
    if no_gt:
        convert_nx_npz_all(countries)
    else:
        convert_nx_gt_all(countries)
        convert_gt_npz_all(countries)
    if len(countries) >= 2:
        merge_all(countries)

def init_stage0(overpass_url,countries,no_gt,url=None,show=False):
    from limic.download import common
    countries, url = common(countries,url,show)
    fill_all(overpass_url,countries)
    extract_cache_all(overpass_url,countries)
    convert_merge_all(countries,no_gt)

def init_stage1_cache(overpass_url,countries,no_gt,url=None,show=False):
    from limic.download import download_cache, common
    countries, url = common(countries,url,show)
    download_cache(countries,url=url)
    extract_cache_all(overpass_url,countries)
    convert_merge_all(countries,no_gt)

def init_stage1_osm(countries,no_gt,url=None,show=False,conserve_mem=False):
    from limic.download import download_osm, common
    download_osm(countries,url=url,show=show)
    countries, url = common(countries,url,show,osm=True)
    extract_osm_all(countries,conserve_mem=conserve_mem)
    convert_merge_all(countries,no_gt)

def init_stage2(countries,no_gt,url=None,show=False):
    from limic.download import download_graph, common
    countries, url = common(countries,url,show)
    download_graph("nx",countries=countries,url=url)
    convert_merge_all(countries,no_gt)

def init_stage3(countries,no_gt,url=None,show=False):
    from limic.download import download_graph, common
    countries, url = common(countries,url,show)
    for suffix in "nx", "npz", "gt":
        download_graph(suffix,countries,url=url)
    if len(countries) >= 2:
        merge_all(countries)



CACHE = (("file_name_in",),{'type':str,'help':"read map data from cache file CACHE",'metavar':'CACHE'})
LIM = (("file_name_in",),{'type':str,'help':"read map data from LiMiC file FILE",'metavar':'LIM'})
OSM = (("file_name_in",),{'type':str,'help':"read map data from OpenStreetMap file OSM",'metavar':'OSM'})
OUT = (("file_name_out",),{'type':str,'help':"save extracted graph to file GRAPH",'metavar':'GRAPH'})
EPS = (("-e","--epsilon"),{'type':float,'dest':'eps','default':0.01,'help':"set minimum percentage of  linear infrastructure segment to consider non-logical intersections to EPS",'metavar':'EPS'})
AROUND = (("-r","--around"),{'type':int,'dest':'around','default':1000,'help':"set maximum free flying distance in m around linear infrastructure nodes to AROUND",'metavar':'AROUND'})
SAFEDIST = (("-s","--safe-dist"),{'type':int,'dest':'safe_dist','default':100,'help':"set minimum safe distance from blacklisted areas and structures to SAFE",'metavar':'SAFE'})
PENALIZE = (("-p","--penalize"),{'type':int,'default':20,'dest':'penalize','help':"penalize free flying disrances by factor PENALTY",'metavar':'PENALTY'})
WHITE = (("--white",),{'type':str,'dest':'white','default':'{"power":"line"}','help':"linear infrastructure to include (default: {'power':'line'})",'metavar':'WHITELIST'})
BLACK = (("--black",),{'type':str,'dest':'black','default':'{"power":"substation"}','help':"areas and structures to avoid (default: {'power':'substation'})",'metavar':'BLACKLIST'})
CONSERVEMEM = (("-c","--conserve-memory"),{'action':'store_true','dest':'conserve_mem','default':False,'help':"lower memory usage but higher runtime"})

CONFIG = [
    ("cache",{'args':[
        (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL",'metavar':'URL'}),
        (("-a","--area"),{'type':str,'dest':'area','help':"set area to extract to AREA",'metavar':'AREA'}),
        AROUND,EPS,SAFEDIST,PENALIZE,CACHE,OUT
    ]}),
    ("osm",{'args':[BLACK,WHITE,CONSERVEMEM,AROUND,EPS,SAFEDIST,PENALIZE,OSM,OUT]}),
    ("osm_pre",{'args':[BLACK,WHITE,CONSERVEMEM,OSM,OUT]}),
    ("osm_post",{'args':[AROUND,EPS,SAFEDIST,PENALIZE,LIM,OUT]})
    #("esy",{'args':[OSM,OUT]})
]

def prune_incomplete(g):
    to_prune = set()
    for k in g.nodes():
        gk = g[k]
        for l in g.neighbors(k):
            if k < l:
                gl = g[l]
                for m in g.neighbors(l):
                    if g.has_edge(k,m) and gk[l]['weight'] > gk[m]['weight']+gl[m]['weight']:
                        to_prune.add((k,l))
    for u,v in to_prune:
        if g.has_edge(u,v):
            g.remove_edge(u,v)

def prune_complete(g):
    from networkx import astar_path_length
    to_prune = set()
    for k in g.nodes():
        for l in g.neighbors(k):
            if k < l and g[k][l]['weight'] > astar_path_length(g,k,l):
                to_prune.add((k,l))
    for u,v in to_prune:
        if g.has_edge(u,v):
            g.remove_edge(u,v)

def build_edges(g,find_all_neighbours,around,eps,safe_dist,penalize):
    from limic.overpass import pylon
    from limic.util import distance
    #count = 0
    neighbours2intersection = {}
    minusid = [0]
    latlon2id = {}
    for tower in list(g.nodes()):
        #count += 1
        #if verbosity >= 2: print(count,"of",total)
        for neighbour in find_all_neighbours(tower,around,eps,safe_dist,minusid,latlon2id,penalize):
            if neighbour.tower in g or neighbour.tower.id < 0:
                #if verbosity >= 2: print("adding",neighbour)
                g.add_edge(tower,pylon(neighbour.tower.id,neighbour.tower.latlon),weight=neighbour.dist,type=neighbour.air)
                if neighbour.tower.id < 0:
                    for intersection in neighbours2intersection.setdefault(neighbour.tower.neighbours,[]):
                        if intersection.latlon != neighbour.tower.latlon:
                            #if verbosity >= 2: print("double intersection:",intersection.latlon,neighbour.tower.latlon)
                            g.add_edge(intersection,neighbour.tower,weight=distance(intersection.latlon,neighbour.tower.latlon),type=False)
                    neighbours2intersection[neighbour.tower.neighbours].append(neighbour.tower)


def extract_cache(file_name_in,file_name_out,overpass_url,area=None,around=1000,eps=0.01,safe_dist=100,penalize=20):
    from limic.overpass import distance, find_all_neighbours, is_safe, set_server, pylon, region, get_towers_by_area
    from limic.util import start, end, file_size, status, load_pickled, save_pickled, replace, check_overwrite
    from networkx import Graph
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading",file_name_in)
    region.backend._cache = load_pickled(file_name_in)
    len_cache = len(region.backend._cache)
    end('')
    file_size(file_name_in)
    if not area:
        area = file_name_in.split(".")[1]
    start("Querying overpass for",area)
    set_server(overpass_url)
    towers = get_towers_by_area(area)
    end()
    start("Building safe nodes")
    g=Graph()
    for tower in towers:
        if is_safe(tower,safe_dist):
            g.add_node(tower)
#        else:
#        if verbosity >= 2: print("NOT safe!")
    end('')
    total = len(g.nodes())
    status(total)
    start("Building edges")
    build_edges(g,find_all_neighbours,around,eps,safe_dist,penalize)
    end('')
    status(len(g.edges()))
    if len_cache != len(region.backend._cache):
        file_name_tmp = file_name_in+".tmp"
        start("Saving to",file_name_in,"via",file_name_tmp)
        save_pickled(file_name_tmp,region.backend._cache)
        replace(file_name_tmp,file_name_in)
        end('')
        file_size(file_name_in)
    common_end(g,file_name_out)
        
def common_end(g,file_name_out,id2types=None):
    from limic.util import start, end, status, file_size, save_pickled
    from networkx import relabel_nodes
    start("Prune redundant edges (incomplete)")
    prune_incomplete(g)
    end('')
    status(len(g.edges()))
    start("Prune redundant edges (complete)")
    prune_complete(g)
    end('')
    status(len(g.edges()))
    start("Cleaning up graph")
    relabel = dict(map(lambda tower:(tower,(tower.id,tower.latlon[0],tower.latlon[1])),g.nodes()))
    relabel_nodes(g,relabel,copy=False)
    if id2types:
        for u,v,d in g.edges(data=True):
            if u[0] >= 0:
                current = u
            elif v[0] >= 0:
                current = v
            else:
                todo = [u]
                done = set()
                current = None
                while todo:
                    current = todo.pop()
                    if current[0] >= 0:
                        break
                    if current in done:
                        continue
                    done.add(current)
                    for n in g.neighbors(current):
                        if not n in done:
                            todo.append(n)
                assert(current is not None)
            d['type'] = -1 if d['type'] else id2types[current[0]]
    else:
        for u,v,d in g.edges(data=True):
            d['type'] = -1 if d['type'] else 0
    end()
    start("Saving graph to",file_name_out)
    save_pickled(file_name_out,g)
    end('')
    file_size(file_name_out)

def is_safe(m_nodes,id2substations,id2elem,tower,around):
    from shapely.geometry import Point, Polygon
    nodes = [] if m_nodes[1] is None else list(map(lambda x:m_nodes[0][x],m_nodes[1].query_ball_point(m_nodes[2].transform(*tower.latlon),r=around)))
    if not nodes:
        return True
    p = Point(*tower.latlon)
    substations = []
    for node in nodes:
        for substation in id2substations[node.id]:
            if not substation in substations:
                substations.append(substation)
    for substation in substations:
        if len(substation) < 3:
            continue
        poly = Polygon(list(map(lambda x:id2elem[x].latlon,substation)))
        if p.within(poly):
            return False
    return True

def find_neighbours(tower,id2elem,id2lines):
    towers = set()
    if tower.id < 0:
        towers.update(tower.neighbours)
        return towers
    for line in id2lines[tower.id]:
        index = line.index(tower.id)
        if index > 0:
            towers.add(line[index-1])
        if index+1 < len(line):
            towers.add(line[index+1])
    return list(map(lambda id:id2elem[id],towers))

def find_segments(tower,around,id2elem,id2lines,neighbourhood):
    segments = []
    lines = []
    for other in neighbourhood:
        for line in id2lines[other.id]:
            if not line in lines:
                lines.append(line)
    for line in lines:
        for i in range(1,len(line)):
            segments.append((line[i-1],line[i]))
    return list(map(lambda x:(id2elem[x[0]],id2elem[x[1]]),segments))            

def find_all_neighbours(tower,around,eps,safe_dist,minusid,latlon2id,penalize,m_towers,m_nodes,id2elem,id2lines,id2substations,latlon2safe):
    from limic.overpass import dist, dist_towers, intersect, pylon
    from limic.util import distance
    direct_neighbours = find_neighbours(tower,id2elem,id2lines)
    neighbours = dist_towers(direct_neighbours,tower.latlon)
    direct_segments = list(map(lambda neighbour:(neighbour.tower,tower),neighbours))
    #print(tower)
    #print(direct_neighbours)
    #print(neighbours)
    neighbourhood = [] if m_towers[1] is None else list(map(lambda x:m_towers[0][x],m_towers[1].query_ball_point(m_towers[2].transform(*tower.latlon),r=around)))
    for neighbour in map(lambda x:dist(0.,False,x),neighbourhood):
        if not neighbour.tower in direct_neighbours and neighbour.tower.id != tower.id:
            neighbour.dist = distance(neighbour.tower.latlon,tower.latlon)*penalize
            neighbour.air = True
            neighbours.append(neighbour)
            #print(neighbour)
    indirect_segments = find_segments(tower,around,id2elem,id2lines,neighbourhood)
    for indirect_segment in indirect_segments:
        for direct_segment in direct_segments:
            intersection = intersect(direct_segment[0].latlon,direct_segment[1].latlon,indirect_segment[0].latlon,indirect_segment[1].latlon,eps)
            if intersection:
                #if verbosity >= 3:
                #    print("found intersection",direct_segment,indirect_segment,intersection)
                #    print("node(id:"+(",".join(map(lambda t:str(t.id),[direct_segment[0],direct_segment[1],indirect_segment[0],indirect_segment[1]])))+");out;")
                if not intersection in latlon2id:
                    minusid[0] -= 1
                    latlon2id[intersection] = minusid[0]
                neighbours.append(dist(distance(intersection,tower.latlon),False,pylon(latlon2id[intersection], intersection,neighbours=indirect_segment)))
    safe_neighbours = []
    for neighbour in neighbours:
        safe = latlon2safe.get(neighbour.tower.latlon,None)
        if safe == None:
            safe = is_safe(m_nodes,id2substations,id2elem,neighbour.tower,safe_dist)
            latlon2safe[neighbour.tower.latlon] = safe
        if safe:
            safe_neighbours.append(neighbour)
        #else:
            #if verbosity >= 4: print("unsafe and skipped:",neighbour.tower)
    return safe_neighbours

def extract_osm(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20,white="{'power':'line'}",black="{'power':'substation'}",conserve_mem=False):
    from limic.util import check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    white,black = list(eval(white).items()), list(eval(black).items())
    lim = osm_pre(file_name_in,white,black,conserve_mem)
    osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20)

def extract_osm_pre(file_name_in,file_name_out,white="{'power':'line'}",black="{'power':'substation'}",conserve_mem=False):
    from limic.util import start, end, save_pickled, file_size, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    white,black = list(eval(white).items()), list(eval(black).items())
    lim = osm_pre(file_name_in,white,black,conserve_mem)
    start("Saving data to",file_name_out)
    save_pickled(file_name_out,lim)
    end('')
    file_size(file_name_out)

def extract_osm_post(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20):
    from limic.util import start, end, file_size, load_pickled, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading filtered OSM data from",file_name_in)
    lim = load_pickled(file_name_in)
    end('')
    file_size(file_name_in)
    osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20)

def osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20):
    from limic.util import start, end, status, file_size, load_pickled, haversine_distance
    #from limic.mtree import MTree
    from scipy.spatial import cKDTree as KDTree
    from networkx import Graph
    from pyproj import CRS, Transformer
    lines, substations, towers, nodes, id2elem, id2lines, id2substations, id2types = lim
    #start("Loading filtered OSM data from",file_name_in)
    #lines, substations, towers, nodes, id2elem, id2lines, id2substations = load_pickled(file_name_in)
    #end()
    start("Building KD-trees from nodes")
    #def distance(x,y):
    #    return haversine_distance(longx=x.latlon[1],latx=x.latlon[0],longy=y.latlon[1],laty=y.latlon[0])
    #m_towers = MTree(d=distance)
    #m_towers.add_all(towers)
    #m_nodes = MTree(d=distance)
    #m_nodes.add_all(nodes)
    crs_4326 = CRS("WGS84")
    crs_proj = CRS("EPSG:28992")
    transformer = Transformer.from_crs(crs_4326, crs_proj)
    towers_tree = KDTree(list(map(lambda x:transformer.transform(x.latlon[0],x.latlon[1]),towers))) if towers else None
    m_towers = (towers,towers_tree,transformer)
    nodes_tree = KDTree(list(map(lambda x:transformer.transform(x.latlon[0],x.latlon[1]),nodes))) if nodes else None
    m_nodes = (nodes,nodes_tree,transformer)
    end()
    #print(m_towers.size)
    #print(m_nodes.size)
    #print("node(id:"+(",".join(list(map(lambda x:str(x.obj.id) if x else "",m_towers.search(towers[0],r=3000)))))+");out;")
    from limic.overpass import pylon
    #find_segments(pylon(1847250054,(56.3973359,8.7254833)),around,m_towers,id2elem,id2lines)
    start("Building safe nodes")
    latlon2safe = {}
    g=Graph()
    for tower in towers:
        safe = is_safe(m_nodes,id2substations,id2elem,tower,safe_dist)
        latlon2safe[tower.latlon] = safe
        if safe:
            g.add_node(tower)
#        else:
#            if verbosity >= 2: print("NOT safe!")
    end('')
    total = len(g.nodes())
    status(total)
    start("Building edges")
    def fan(tower,around,eps,safe_dist,minusid,latlon2id,penalize):
        return find_all_neighbours(tower,around,eps,safe_dist,minusid,latlon2id,penalize,m_towers,m_nodes,id2elem,id2lines,id2substations,latlon2safe)
    build_edges(g,fan,around,eps,safe_dist,penalize)
    end('')
    status(len(g.edges()))
    common_end(g,file_name_out,id2types)

def osm_pre(file_name_in,white=[("power","line")],black=[("power","substation")],conserve_mem=False):
    from limic.util import start, end, status, file_size, save_pickled
    from xml.sax import parse
    from xml.sax.handler import ContentHandler
    from bz2 import open as bopen
    from limic.overpass import pylon
    start("Extracting ways for power lines from",file_name_in)
    class ways_handler(ContentHandler):
        def __init__(self):
            self.nodes = None
            self.power = None
            self.lines = []
            self.substations = []
            self.towers = []
            self.type = None
            self.id2types = {}
        def startElement(self, name, attrs):
            if name == 'way':
                self.nodes = []
                self.power = None
            elif name == 'tag':
                kv = (attrs['k'],attrs['v'])
                if kv in white:
                    self.power = "white"
                    self.type = white.index(kv)
                elif kv in black:
                    self.power = "black"
            elif name == 'nd':
                self.nodes.append(int(attrs['ref']))
            elif name == 'node':
                if not conserve_mem:
                    self.towers.append(pylon(int(attrs['id']),(float(attrs['lat']),float(attrs['lon']))))
        def endElement(self, name):
            if name == 'way':
                if self.power == 'white':
                    for node in self.nodes:
                        self.id2types[node] = self.type
                    self.lines.append(self.nodes)
                elif self.power == 'black':
                    self.substations.append(self.nodes)
    f = bopen(file_name_in,"rb") if file_name_in[-4:] == ".bz2" else open(file_name_in,"rb")
    ways = ways_handler()
    parse(f, ways)
    f.close()
    lines, substations, id2types = ways.lines, ways.substations, ways.id2types
    end('')
    status(len(lines),end='   ')
    status(len(substations))
    id2lines, id2substations = {}, {}
    for line in lines:
        for node in line:
            id2lines.setdefault(node,[]).append(line)
    for substation in substations:
        for node in substation:
            id2substations.setdefault(node,[]).append(substation)
    if conserve_mem:
        del ways
        start("Extracting nodes for white and black")
        class nodes_handler(ContentHandler):
            def __init__(self):
                self.towers = []
                self.nodes = []
                self.id2elem = {}
            def startElement(self, name, attrs):
                if name == 'node':
                    i = int(attrs['id'])
                    if i in id2lines:
                        elem = pylon(i,(float(attrs['lat']),float(attrs['lon'])))
                        self.towers.append(elem)
                        self.id2elem[i] = elem
                    if i in id2substations:
                        elem = pylon(i,(float(attrs['lat']),float(attrs['lon'])))
                        self.nodes.append(elem)
                        self.id2elem[i] = elem
        f = bopen(file_name_in,"rb") if file_name_in[-4:] == ".bz2" else open(file_name_in,"rb")
        ns = nodes_handler()
        parse(f, ns)
        f.close()
        towers, nodes, id2elem = ns.towers, ns.nodes, ns.id2elem
        del ns
        end('')
        status(len(towers),end='   ')
        status(len(nodes))
    else:
        start("Extracting nodes for white and black")
        towers = []
        nodes = []
        id2elem = {}
        for tower in ways.towers:
            both = False
            if tower.id in id2lines:
                id2elem[tower.id] = tower
                towers.append(tower)
            if tower.id in id2substations:
                id2elem[tower.id] = tower
                nodes.append(tower)
        del ways
        end()
    start("Checking that all nodes were extracted for",file_name_in)
    assert(set(id2lines.keys())==set(map(lambda x:x.id,towers)))
    assert(set(id2substations.keys())==set(map(lambda x:x.id,nodes)))
    end()
    return lines,substations,towers,nodes,id2elem,id2lines,id2substations,id2types



CONFIG = [
    (("-a","--area"),{'type':str,'dest':'area','help':"set area to extract to AREA",'metavar':'AREA'}),
    (("file_name",),{'type':str,'nargs':'?','help':"save map data to cache file CACHE",'metavar':'CACHE'}),
    (("-w","--max-workers"),{'type':int,'dest':'max_workers','help':"set maximum number of parallel workers to WORKERS",'metavar':'WORKERS'}),
    (("-r","--around"),{'type':int,'dest':'around','default':1000,'help':"set maximum free flying distance in m around power towers to AROUND",'metavar':'AROUND'}),
    (("-e","--epsilon"),{'type':float,'dest':'eps','default':0.01,'help':"set minimum percentage of power line segment to consider non-logical intersections to EPS",'metavar':'EPS'}),
    (("-s","--safe-dist"),{'type':int,'dest':'safe_dist','default':100,'help':"set minimum safe distance from transformer stations and other perilous structures to SAFE",'metavar':'SAFE'}),
    (("-p","--penalize"),{'type':int,'default':20,'dest':'penalize','help':"penalize air time by factor PENALTY",'metavar':'PENALTY'}),
    (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL",'metavar':'URL'})
]

def cache_tower(tower, around, eps, safe_dist, penalize):
    from limic.overpass import find_all_neighbours
    find_all_neighbours(tower,around,eps,safe_dist,[0],{},penalize)

def fill(overpass_url,file_name=None,area=None,around=1000,eps=0.01,safe_dist=100,penalize=20,max_workers=None):
    from limic.overpass import set_server, pylon, region, get_towers_by_area
    from limic.util import start, end, file_size, status, load_pickled, save_pickled, options, replace, options
    from networkx import Graph, relabel_nodes
    from os import cpu_count
    from os.path import exists
    from concurrent.futures import ThreadPoolExecutor, wait
    from signal import signal, SIGINT
    if not area and not file_name:
        if options.parser:
            options.parser.error("specify at least one of --area or CACHE")
        else:
            status("ERROR: specify at least area or cache name!")
            from sys import exit
            exit(-1)
    if not area:
        area = file_name.split(".")[1]
    if not file_name:
        file_name = "cache."+area
    if not max_workers:
        max_workers = cpu_count()*4
    start("Number of workers")
    status(max_workers)
    if exists(file_name):
        start("Loading",file_name)
        region.backend._cache = load_pickled(file_name)
        end('')
        file_size(file_name)
    len_cache = len(region.backend._cache)
    start("Querying overpass for",area)
    set_server(overpass_url)
    towers = get_towers_by_area(area)
    end()
    fs = []
    executor = ThreadPoolExecutor(max_workers=max_workers)
    interrupt = 0
    def shutdown(sig,frame):
        nonlocal interrupt
        interrupt += 1
        print("Shutting down ...")
        for f in fs:
            f.cancel()
        print("Cancelled all futures ...")
        running = len(fs)
        total = running
        while running:
            print("Waiting for",running,"processes to shut down ...")
            wait(fs,timeout=60)
            running = sum(0 if f.done() else 1 for f in fs)
        if len_cache != len(region.backend._cache):
            file_name_tmp = file_name+"."+str(interrupt)
            start("Emergency saving to",file_name_tmp)
            save_pickled(file_name_tmp,region.backend._cache)
            end('')
            file_size(file_name_tmp)
    signal(SIGINT,shutdown)
    options.failed = True
    while options.failed:
        options.failed = False
        for tower in towers:
            fs.append(executor.submit(cache_tower, tower, around, eps, safe_dist, penalize))
        running = len(fs)
        total = running
        while running:
            print("Waiting for",running,"out of",total,"processes ...")
            wait(fs,timeout=60)
            running = sum(0 if f.done() else 1 for f in fs)
    if len_cache != len(region.backend._cache):
        file_name_tmp = file_name+".tmp"
        start("Saving to",file_name,"via",file_name_tmp)
        save_pickled(file_name_tmp,region.backend._cache)
        replace(file_name_tmp,file_name)
        end('')
        file_size(file_name)



OVERPASS = (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL (default: http://caracal.imada.sdu.dk/)",'metavar':'URL'})
IN = (("file_name_in",),{'type':str,'help':"read map data from cache file CACHE",'metavar':'CACHE'})
OUT = (("file_name_out",),{'type':str,'help':"save extracted graph to file GRAPH",'metavar':'GRAPH'})    
COORDS = (("polygon",),{'type':str,'nargs':'+','help':"coordinate pairs that describe geometry",'metavar':'LAT LON'})

CONFIG = [
    ("nx",{'help':"prune NX graph",'args':[OVERPASS,IN,OUT,COORDS]}),
    ("gt",{'help':"prune GT graph",'args':[OVERPASS,IN,OUT,COORDS]}),
    ("npz",{'help':"prune NPZ graph",'args':[OVERPASS,IN,OUT,COORDS]}),
]

def prune_ids_nx(g,delete_ids):
    to_delete = []
    for n in g.nodes():
        if n[0] in delete_ids:
            to_delete.append(n)
    for n in to_delete:
        g.remove_node(n)
    return g

def prune_nx(file_name_in,file_name_out,polygon,overpass_url):
    from limic.util import start, end, file_size, status, save_pickled, load_pickled, check_overwrite
    from limic.overpass import nodes_in_geometry, set_server
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading from",file_name_in)
    g = load_pickled(file_name_in)
    end('')
    file_size(file_name_in)
    start("Query server for nodes in polygon")
    set_server(overpass_url)
    nodes = nodes_in_geometry(zip(polygon[::2], polygon[1::2]))
    end('')
    status(len(nodes))
    start("Pruning graph")
    h = prune_ids_nx(g,nodes)
    end()
    start("Saving to",file_name_out)
    save_pickled(file_name_out,h)
    end('')
    file_size(file_name_out)
    
def prune_ids_gt(g,delete_ids):
    to_delete = []
    ids = g.get_vertices([g.vp.id])
    for v in ids:
        if v[1] in delete_ids:
            to_delete.append(g.vertex(v[0]))
    to_delete.reverse()
    for v in to_delete:
        g.remove_vertex(v)
    return g

def prune_gt(file_name_in,file_name_out,polygon,overpass_url):
    from limic.util import start, end, file_size, status, save_gt, load_gt, check_overwrite
    from limic.overpass import nodes_in_geometry, set_server
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading from",file_name_in)
    g = load_gt(file_name_in)
    end('')
    file_size(file_name_in)
    start("Query server for nodes in polygon")
    set_server(overpass_url)
    nodes = nodes_in_geometry(zip(polygon[::2], polygon[1::2]))
    end('')
    status(len(nodes))
    start("Pruning graph")
    h = prune_ids_gt(g,nodes)
    end()
    start("Saving to",file_name_out)
    save_gt(file_name_out,h)
    end('')
    file_size(file_name_out)

def prune_ids_npz(g,delete_ids):
    from numpy import array, int64, int32, float64, float32
    g_ids = g['ids']
    g_lat = g['lat']
    g_long = g['long']
    g_id2edges = g['id2edges']
    g_edges_weight = g['edges_weight']
    g_edges_neighbor = g['edges_neighbor']
    h_ids = []
    h_lat = []
    h_long = []
    h_id2edges = []
    h_edges_weight = []
    h_edges_neighbor = []
    g2h = {}
    h_index = 0
    for g_index in range(len(g_ids)):
        if g_ids[g_index] in delete_ids:
            continue
        g2h[g_index] = h_index
        h_ids.append(g_ids[g_index])
        h_lat.append(g_lat[g_index])
        h_long.append(g_long[g_index])
        h_index += 1
    h_edge_index = 0
    for g_index in range(len(g_ids)):
        if g_index not in g2h:
            continue
        h_id2edges.append(h_edge_index)
        g_edge_index = g_id2edges[g_index]
        g_edge_index_next = g_id2edges[g_index+1]
        while g_edge_index < g_edge_index_next:
            neighbor = g_edges_neighbor[g_edge_index]
            if neighbor in g2h:
                h_edges_weight.append(g_edges_weight[g_edge_index])
                h_edges_neighbor.append(g2h[neighbor])
                h_edge_index += 1
            g_edge_index += 1
            assert(h_edge_index == len(h_edges_weight))
    h_id2edges.append(h_edge_index)
    h = {}
    h['ids'] = array(h_ids,dtype=int64)
    h['lat'] = array(h_lat,dtype=float64)
    h['long'] = array(h_long,dtype=float64)
    h['id2edges'] = array(h_id2edges,dtype=int32)
    h['edges_weight'] = array(h_edges_weight,dtype=float32)
    h['edges_neighbor'] = array(h_edges_neighbor,dtype=int32)
    return h

def prune_npz(file_name_in,file_name_out,polygon,overpass_url):
    from limic.util import start, end, file_size, status, save_npz, load_npz, check_overwrite
    from limic.overpass import nodes_in_geometry, set_server
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading from",file_name_in)
    g = load_npz(file_name_in)
    end('')
    file_size(file_name_in)
    start("Query server for nodes in polygon")
    set_server(overpass_url)
    nodes = nodes_in_geometry(zip(polygon[::2], polygon[1::2]))
    end('')
    status(len(nodes))
    start("Pruning graph")
    h = prune_ids_npz(g,nodes)
    end()
    start("Saving to",file_name_out)
    save_npz(file_name_out,h)
    end('')
    file_size(file_name_out)



# arguments
def parse_config(config, parser, prefix):
    subs = None
    for a,b in config:
        if isinstance(a,str):
            if not subs:
                subs = parser.add_subparsers(dest='command',required=True)
            args = b.pop('args') if 'args' in b else None
            sub = subs.add_parser(a,**b)
            if not prefix:
                sub.set_defaults(mod="limic."+a)
            sub.set_defaults(func="_".join(prefix+[a]))
            if args:
                parse_config(args, sub, prefix+[a])
        else:
            parser.add_argument(*a, **b)

# global options
from argparse import Namespace
options = Namespace()
options.verbosity = 1
options.md5sum = False
options.overwrite = False
options.parser = None
options.disk_cache = None
options.failed = False
def set_option(k,v):
    if k is not None:
        vars(options)[k] = v

# timing and informational messages
import time
started = 0
def start(*msg):
    if options.verbosity > 0:
        global started
        if msg:
            print(" ".join(map(str,msg)).ljust(60),"... ",end='',flush=True)
        started = time.time()
def end(end='\n'):
    if options.verbosity > 0:
        global started
        print("%.3f seconds   " % (time.time()-started),end=end,flush=True)
        started = time.time()
def file_size(file_name):
    from os import stat
    if options.verbosity > 0:
        print("%.0fK" % (stat(file_name).st_size/1024))
def status(msg,end='\n'):
    if options.verbosity > 0:
        print(msg,end=end)


# distance between GPS positions
from math import radians, cos, sin, asin, sqrt
def haversine_distance(longx,latx,longy,laty):
    lon1, lat1, lon2, lat2 = radians(longx), radians(latx), radians(longy), radians(laty)
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    m = 6371008.7714*c
    return m
def distance(x,y):
    return haversine_distance(longx=x[1],latx=x[0],longy=y[1],laty=y[0])

# working with different graph formats
def nx_is_equal(g,h,isomorphic=False):
    from networkx.algorithms.isomorphism import is_isomorphic
    if isomorphic:
        return is_isomorphic(g,h,lambda ga,ha:ga==ha)
    g_edges, h_edges = g.edges(), h.edges()
    return all(h_edge in g_edges for h_edge in h_edges) and all(g_edge in h_edges for g_edge in g_edges)
def locate_by_id(ids,source_id=None,target_id=None,length=None):
    if not length:
        length = len(ids)
    from random import randrange
    source = None if source_id else randrange(length)
    target = None if target_id else randrange(length)
    for i in range(length):
        if source:
            if target:
                return source,target
            if ids[i] == target_id:
                target = i
                continue
        else:
            if target:
                if ids[i] == source_id:
                    source = i
                    continue
            else:
                if ids[i] == source_id:
                    source = i
                if ids[i] == target_id:
                    target = i
    assert False, "cannot resolve all ids: %s,%s -> %s,%s" % tuple(map(str,(source_id,target_id,source,target)))

# loading and saving different formats
def check_overwrite(file_name_in,file_name_out):
    return check_overwrites([file_name_in],file_name_out)
def check_overwrites(file_names,file_name_out):
    if not options.md5sum or options.overwrite:
        return True
    from os.path import exists, getmtime
    if not exists(file_name_out):
        return True
    out_mtime = getmtime(file_name_out)
    for file_name_in in file_names:
        if getmtime(file_name_in) > out_mtime:
            return True
    if not check_md5_internal(file_name_out) is None:
        return True
    status("INFO(overwrite): SKIPPING generation of "+file_name_out+" as it is newer than "+(" ".join(file_names)))
    return False
def replace(file_name_in,file_name_out):
    from os import replace
    from os.path import exists
    replace(file_name_in,file_name_out)
    if exists(file_name_in+".md5"):
        replace(file_name_in+".md5",file_name_out+".md5")
def md5file(file_name):
    from hashlib import md5
    from os.path import exists
    if not exists(file_name):
        from sys import exit
        status("ERROR(md5file): could not find "+file_name)
        exit(-1)
    f = open(file_name,"rb")
    sum = md5(f.read()).hexdigest()
    f.close()
    return sum
def check_md5_internal(file_name):
    from os.path import exists
    if not exists(file_name+".md5"):
        return "notexists"
    f = open(file_name+".md5")
    md5_1 = f.read().split()[0]
    f.close()
    md5_2 = md5file(file_name)
    if md5_1 != md5_2:
        return "notequal"
    return None
def check_md5(file_name):
    if options.md5sum:
        res = check_md5_internal
        if res == "notexists":
            from sys import exit
            status("ERROR(md5file): could not find "+file_name+".md5")
            exit(-1)
        if res == "notequal":
            from sys import exit
            status("ERROR(md5): "+md5_1+" vs "+md5_2)
            exit(-1)
def save_md5(file_name):
    if options.md5sum:
        f = open(file_name+".md5","wt")
        f.write(md5file(file_name)+" "+file_name)
        f.close()
def save_pickled(file_name,g,compression="gzip",protocol=4):
    from pickle import dump
    if compression:
        from gzip import open as gopen
        f = gopen(file_name,"wb")
    else:
        f = open(file_name,"wb")
    dump(g,f,protocol=protocol)
    f.close()
    save_md5(file_name)
def load_pickled(file_name,compression="gzip"):
    check_md5(file_name)
    from pickle import Unpickler
    class RenameUnpickler(Unpickler):
        def find_class(self, module, name):
            renamed_module = module
            if module == "overpass":
                renamed_module = "limic.overpass"
            return super(RenameUnpickler, self).find_class(renamed_module, name)
    #from pickle import load
    def load(f):
        return RenameUnpickler(f).load()
    if compression:
        from gzip import open as gopen
        f = gopen(file_name,"rb")
    else:
        f = open(file_name,"rb")
    g = load(f)
    f.close()
    return g
def save_gt(file_name,g):
    g.save(file_name,fmt="gt")
    save_md5(file_name)
def load_gt(file_name):
    check_md5(file_name)
    from graph_tool import Graph
    g = Graph()
    g.load(file_name,fmt="gt")
    return g
def load_npz(file_name):
    check_md5(file_name)
    from numpy import load
    return load(file_name)
def save_npz(file_name,g):
    from numpy import savez
    savez(file_name,**g)
    save_md5(file_name)
def save_path(costpath,out_file=None,visualize=False):
    from sys import stdout
    if visualize and not out_file:
        status("WARNING (use --output to specify HTML file)   ",end='')
        visualize = False
    cost,path = costpath
    if visualize:
        from folium import Map, Marker, Icon, PolyLine
        from folium.plugins import BeautifyIcon
        from binascii import hexlify
        from webbrowser import open as wopen
        from pathlib import Path
        from math import log2
        min_lat = min_long = float('inf')
        max_lat = max_long = -float('inf')
        for x in path:
            if x[4] < min_lat: min_lat = x[4]
            if x[5] < min_long: min_long = x[5]
            if x[4] > max_lat: max_lat = x[4]
            if x[5] > max_long: max_long = x[5]
        m = Map()
        m.fit_bounds([(min_lat,min_long),(max_lat,max_long)])
        start_color = (63,255,63)
        end_color = (63,63,255)
        diff_color = tuple(map(lambda x:x[0]-x[1],zip(end_color,start_color)))
        length = float(len(path))
        for i in range(len(path)):
            x = path[i]
            background_color = "#"+hexlify(bytes(map(lambda x:int(x[0]+i*x[1]/length),zip(start_color,diff_color)))).decode('utf8')
            border_color = "#"+hexlify(bytes(map(
                lambda x:int((x[0]+i*x[1]/length)/2),zip(start_color,diff_color)))).decode('utf8')
            line_color = background_color
            line_weight = 6
            icon = None
            iconStyle = ""
            if x[3] < 0:
                icon = 'times'
                background_color = "#ffff3f"
                border_color = "#7f1f1f"
            elif x[2]:
                icon = 'flash'
                background_color = "#ff0000"
                border_color = "#7f0000"
                line_color = background_color
                line_weight = 10
            elif i+1 < len(path) and path[i+1][2]:
                icon = 'flash'
                background_color = "#ff0000"
                border_color = "#7f0000"
            elif i == 0 or i+1 == len(path):
                icon = 'flash'
            else:
                icon = 'none'
                iconStyle = "opacity: 0.1;"
            if i > 0 and cost is not None:
                PolyLine([path[i-1][4:6],path[i][4:6]],color=line_color,opacity=0.4,weight=line_weight).add_to(m)
            if icon:
                Marker(x[4:6],icon=BeautifyIcon(icon=icon,
                                                iconStyle=iconStyle,
                                            borderColor=border_color,
                                            backgroundColor=background_color),
                   popup=("cost: %.1fm, dist: %.1fm, air: %r, id: %d" % x[:4])).add_to(m)
        m.save(out_file)
        wopen(Path(out_file).resolve().as_uri(),new=2)
    else:
        file = open(out_file,"w") if out_file else stdout
        print("node(id:"+",".join(map(lambda x:str(x[3]),path))+");out;",file=file)



from limic.route import CONFIG as ROUTE_CONFIG
from limic.init import CONFIG as INIT_CONFIG
from limic.download import CONFIG as DOWNLOAD_CONFIG
from limic.merge import CONFIG as MERGE_CONFIG
from limic.convert import CONFIG as CONVERT_CONFIG
from limic.prune import CONFIG as PRUNE_CONFIG
from limic.select import CONFIG as SELECT_CONFIG
from limic.length import CONFIG as LENGTH_CONFIG
from limic.fill import CONFIG as FILL_CONFIG
from limic.extract import CONFIG as EXTRACT_CONFIG
from limic.render import CONFIG as RENDER_CONFIG
from limic.serve import CONFIG as SERVE_CONFIG
from limic.test import CONFIG as TEST_CONFIG
CONFIG = [
    (("-v","--verbosity"),{'type':int,'default':1,'dest':'verbosity','help':"set verbosity level to LEVEL (default: 1)",'metavar':'LEVEL'}),
    (("-5","--md5"),{'action':'store_true','default':False,'dest':'md5sum','help':"deactivate MD5 checksumming (default: False)"}),
    (("-w","--overwrite"),{'action':'store_true','default':False,'dest':'overwrite','help':"overwrite if destinatone exists (default: False)"}),
    ("init",{'help':"Initialize the data files.",'args':INIT_CONFIG}),
    ("download",{'help':"Download graph, cache, or map files.",'args':DOWNLOAD_CONFIG}),
    ("route",{'help':"Route from one power tower to another.",'args':ROUTE_CONFIG}),
    ("convert",{'help':"Convert graphs/caches.",'args':CONVERT_CONFIG}),
    ("merge",{'help':"Merge graphs/caches.",'args':MERGE_CONFIG}),
    ("prune",{'help':"Prune graph by geometry.",'args':PRUNE_CONFIG}),
    ("select",{'help':"Select area from graph by geometry.",'args':SELECT_CONFIG}),
    ("length",{'help':"Compute total length of non-air edges.",'args':LENGTH_CONFIG}),
    ("fill",{'help':"Fill a cache with map data.",'args':FILL_CONFIG}),
    ("extract",{'help':"Extract NX graph from cache with map data.",'args':EXTRACT_CONFIG}),
    ("render",{'help':"Render graphs to HTML",'args':RENDER_CONFIG}),
    ("serve",{'help':"Serve interactive routing map",'args':SERVE_CONFIG}),
    ("test",{'help':"Extensive self tests.",'args':TEST_CONFIG}),
]

if __name__ == "__main__":
    from argparse import ArgumentParser
    from limic.util import set_option, parse_config
    from importlib import import_module
    parser = ArgumentParser(description="Powerlines as drone highways.")
    parse_config(CONFIG, parser, [])
    args = parser.parse_args()
    set_option('verbosity',args.verbosity)
    set_option('md5sum',not args.md5sum)
    set_option('overwrite',args.overwrite)
    set_option('parser',parser)
    module = import_module(args.mod)
    func = vars(module)[args.func]
    del args.verbosity, args.md5sum, args.overwrite, args.mod, args.func, args.command
    func(**vars(args))



GRAPHS = (("file_names",),{'type':str,'nargs':'+','help':"graphs to merge",'metavar':'GRAPH'})
CACHES = (("file_names",),{'type':str,'nargs':'+','help':"caches to merge",'metavar':'CACHE'})
TARGET = (("file_name_out",),{'type':str,'help':"target file for merge result",'metavar':'TARGET'})

CONFIG = [
    ("nx",{'help':"merge NX graphs",'args':[GRAPHS,TARGET]}),
    ("cache",{'help':"merge caches",'args':[CACHES,TARGET]})
]

def merge_nx(file_names,file_name_out):
    from limic.util import start, end, file_size, status, load_pickled, save_pickled, check_overwrites
    from networkx import Graph
    if not check_overwrites(file_names,file_name_out):
        return
    g = Graph()
    for file_name_in in file_names:
        start("Loading graph from",file_name_in)
        h = load_pickled(file_name_in)
        end('')
        file_size(file_name_in)
        start("Adding",h.number_of_edges(),"edges")
        for from_node, to_node, data in h.edges(data=True):
            g.add_edge(from_node,to_node,**data)
        end('')
        status(g.number_of_edges())
    start("Saving merged graph to",file_name_out)
    save_pickled(file_name_out,g)
    end('')
    file_size(file_name_out)

def merge_cache(file_names,file_name_out):
    from limic.util import start, end, file_size, status, load_pickled, save_pickled, check_overwrites
    if not check_overwrites(file_names,file_name_out):
        return
    g = {}
    for file_name_in in file_names:
        start("Loading cache from",file_name_in)
        h = load_pickled(file_name_in)
        end('')
        file_size(file_name_in)
        start("Adding",len(h),"entries")
        g.update(h)
        end('')
        status(len(g))
    start("Saving merged cache to",file_name_out)
    save_pickled(file_name_out,g)
    end('')
    file_size(file_name_out)
